"""Model base class"""
from datetime import datetime
import numpy as np
from tekinstr.common import TekBase, _get_idn


class Model(TekBase):
    """Base class for a given model

    Attributes:
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, visa):
        super().__init__(visa)
        self._visa.write("HEADER OFF")
        self._visa.write("DESE 255")
        self._visa.write("VERBOSE ON")
        self._idn = _get_idn(visa)

    @property
    def model(self):
        """(str): the model number"""
        return self._idn.model

    @property
    def serial_number(self):
        """(str): the serial number"""
        return self._idn.serial_number

    @property
    def firmware_version(self):
        """(str): the firmware version"""
        return self._idn.firmware_version

    def lock_frontpanel(self, message=None):
        """Lock the front panel controls

        Args:
            message (str): message to display, such as "In use by ..."
        """
        self._visa.write("LOCK ALL")
        if message is not None:
            message = message[:1000]
            self._visa.write(f"MESSAGE:SHOW '{message[:1000]}'")
            self._visa.write("MESSAGE:BOX 0, 0")
            self._visa.write("MESSAGE:STATE ON")

    def unlock_frontpanel(self):
        """Unlock the front panel controls and clear message box if displayed"""
        self._visa.write("UNLOCK ALL")
        self._visa.write("MESSAGE:CLEAR")
        self._visa.write("MESSAGE:STATE OFF")

    def reset(self):
        """Reset the oscilloscope to the factory default settings"""
        self._visa.write("*RST")

    def set_clock(self):
        """Set the date and time to the local machine's time"""
        self._visa.write(f"DATE '{datetime.now().strftime('%Y-%m-%d')}'")
        self._visa.write(f"TIME '{datetime.now().strftime('%H:%M:%S')}'")

    @property
    def time(self):
        """Get date and time"""
        t_str = self._visa.query("TIME?").replace('"', "")
        d_str = self._visa.query("DATE?").replace('"', "")
        return np.datetime64(f"{d_str} {t_str}")

    def __repr__(self):
        return f"<Tektronix {self.model} at {self._visa.resource_name}>"



"""Instrument base class"""
from tekinstr.common import TekBase


class Instrument(TekBase):
    """Base class for a specific type of instrument within a given model

    Attributes:
        instr (Model)
    """

    _kind = "Instrument"

    def __init__(self, instr):
        super().__init__(instr._visa)
        self._instr = instr

    def __init_subclass__(cls, kind):
        cls._kind = kind

    def __get__(self, instance, owner=None):
        return self

    def __repr__(self):
        return f"<{self._instr.model} {self._kind}>"


class InstrumentSubsystem(Instrument, kind="InstrumentSubsystem"):
    """Base class for an instrument subsystem

    Attributes:
        owner (Instrument)
    """

    def __init__(self, owner):
        super().__init__(owner._instr)
        self._owner = owner

    def __repr__(self):
        return f"<{self._instr.model} {self._owner._kind} {self._kind}>"



"""Base Spectrum Analyzer definition"""
import re
from datetime import datetime
import numpy as np
from tekinstr.common import validate
from tekinstr.instrument import Instrument
from waveformDT.waveform import WaveformDT


class SpectrumAnalyzerBase(Instrument, kind="Base Spectrum Analyzer"):
    """Spectrum analyzer instrument

    Attributes:
        instr (Model)
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    def _get_wfmpre(self, inout=""):
        self._visa.write("HEADER ON")
        pre_str = self._visa.query(f"WFM{inout}PRE?").replace(f":WFM{inout}PRE:", "")
        raw_preamble = dict([kv.split(" ", maxsplit=1) for kv in pre_str.split(";")])
        preamble = {}
        for k, v in raw_preamble.items():
            value = v.strip('"')
            try:
                value = int(v)
            except ValueError:
                try:
                    value = float(v)
                except ValueError:
                    pass
            preamble[k] = value
        self._visa.write("HEADER OFF")
        return preamble

    def _get_select(self):
        """Get the current display selection"""
        self._visa.write("HEADER ON")
        select_str = self._visa.query("SELECT?").replace(":SELECT:", "")
        raw_select = dict([kv.split(" ", maxsplit=1) for kv in select_str.split(";")])
        displayed_channels = set()
        control_channel = raw_select.pop("CONTROL", None)
        for k, v in raw_select.items():
            if k.startswith("RF") and v in ["1", "ON"]:
                k = k.split("RF_")[1].replace("MIN", "MIN ").replace("MAX", "MAX ")
                displayed_channels.add(k)
        self._visa.write("HEADER OFF")
        return (displayed_channels, control_channel)

    @property
    def trace(self):
        """value (str): displayed spectrum trace; normal, average, min hold, max hold"""
        return self._get_select()[0]

    @trace.setter
    @validate
    def trace(self, value):
        raw_traces = [value] if isinstance(value, str) else value
        traces = set()
        for raw_tr in raw_traces:
            tr = raw_tr.upper().replace("MIN ", "MIN").replace("MAX ", "MAX")
            traces.add(tr)
        displayed_traces = self.trace
        for tr in displayed_traces - traces:
            self._visa.write(f"SELECT:RF_{tr} OFF")
        for tr in traces - displayed_traces:
            self._visa.write(f"SELECT:RF_{tr} ON")

    @property
    def center_frequency(self):
        """value (float): center frequency in hertz"""
        return float(self._visa.query("RF:FREQUENCY?"))

    @center_frequency.setter
    @validate
    def center_frequency(self, value):
        self._visa.write(f"RF:FREQUENCY {float(value)}")

    @property
    def span(self):
        """value (float): span"""
        return float(self._visa.query("RF:SPAN?"))

    @span.setter
    @validate
    def span(self, value):
        self._visa.write(f"RF:SPAN {float(value)}")

    @property
    def rbw_mode(self):
        """value (str): resolution bandwidth (rbw) mode; AUTO or MANUAL"""
        return self._visa.query("RF:RBW:MODE?")

    @rbw_mode.setter
    @validate
    def rbw_mode(self, value):
        self._visa.write(f"RF:RBW:MODE {value}")

    @property
    def rbw(self):
        """value (float): resolution bandwidth"""
        return float(self._visa.query("RF:RBW?"))

    @rbw.setter
    @validate
    def rbw(self, value):
        self._visa.write(f"RF:RBW {value}")

    @property
    def rbw_ratio(self):
        """value (float): ration of span to resolution bandwidth"""
        return float(self._visa.query("RF:SPANRBWRATIO?"))

    @rbw_ratio.setter
    @validate
    def rbw_ratio(self, value):
        self._visa.write(f"RF:SPANRBWRATIO {float(value)}")

    @property
    def ref_level(self):
        """value (float): reference level"""
        return float(self._visa.query("RF:REFLEVEL?"))

    @ref_level.setter
    @validate
    def ref_level(self, value):
        self._visa.write(f"RF:REFLEVEL {value}")

    @property
    def vertical_position(self):
        """value (float): vertical position"""
        return float(self._visa.query("RF:POSITION?"))

    @vertical_position.setter
    @validate
    def vertical_position(self, value):
        self._visa.write(f"RF:POSITION {float(value)}")

    @property
    def vertical_scale(self):
        """value (float): vertical_unit per division"""
        return float(self._visa.query("RF:SCALE?"))

    @vertical_scale.setter
    @validate
    def vertical_scale(self, value):
        self._visa.write(f"RF:SCALE {float(value)}")

    @property
    def vertical_unit(self):
        """value (str): vertical unit; dBm, dBuw, etc."""
        return self._visa.query("RF:UNITS?")

    @vertical_unit.setter
    @validate
    def vertical_unit(self, value):
        self._visa.write(f"RF:UNITS {value}")

    @property
    def window(self):
        """value (str): window function"""
        return self._visa.query("RF:WINDOW?")

    @window.setter
    @validate
    def window(self, value):
        self._visa.write(f"RF:WINDOW {value}")

    @property
    def label(self):
        """value (str): label"""
        return self._visa.query("RF:LABEL?").strip('"')

    @label.setter
    @validate
    def label(self, value):
        self._visa.write(f"RF:LABEL '{value}'")

    def read(self, dB=True, wdt=True):
        """Transfer RF spectrum

        Args:
            dB (bool): scale data in dB if True, otherwise unit is watts
            wdt (bool): return data as WaveformDT, otherwise Numpy ndarray
        Returns:
            (ndarray or WaveformDT)
        """
        self._visa.write("DATA:SOURCE RF_NORMAL")
        preamble = self._get_wfmpre("OUT")
        time = self._visa.query("TIME?")
        date = self._visa.query("DATE?")
        data = self._visa.query_binary_values(
            "CURVE?", is_big_endian=True, container=np.ndarray
        )
        if dB:
            match = re.match("^DB(?P<prefix>[UM])(?P<unit>[AVW])*$", self.vertical_unit)
            prefix = "m" if match.group("prefix") == "M" else "µ"
            unit = "W" if match.group("unit") is None else match.group("unit")
            y_unit = "dB" + prefix + unit
            scale = 10 if unit == "W" else 20
            ref = 1e-3 if prefix == "m" else 1e-6
            data = scale * np.log10(data / ref)
        else:
            y_unit = "W"
        if wdt:
            time = [int(n) for n in time.strip().strip('"').split(":")]
            date = [int(n) for n in date.strip().strip('"').split("-")]
            t0 = datetime(*date, *time)
            dt = preamble["XINCR"]
            data = WaveformDT(data, dt, t0)
            setattr(data, "wf_start_offset", preamble["XZERO"])
            setattr(data, "x_unit", preamble["XUNIT"])
            setattr(data, "y_unit", y_unit)
        return data

    @property
    def clipping(self):
        """(bool): RF input clipping"""
        value = self._visa.query("RF:CLIPPING?")
        return value in ["1", "ON"]



"""Oscilloscope instrument definition"""
from tekinstr.common import validate
from tekinstr.oscilloscope import OscilloscopeBase, ChannelBase, ProbeBase
from tekinstr.mdo3000.trigger import Trigger
from tekinstr.measurement import Measurement


class Oscilloscope(OscilloscopeBase, kind="Oscilloscope"):
    """Oscilloscope instrument

    Attributes:
        instr (Model)
        n_channels (int): number of channels
    """

    def __init__(self, instr, n_channels):
        super().__init__(instr)
        self._channels = [Channel(self, x) for x in range(1, n_channels + 1)]
        for x, ch in enumerate(self._channels, 1):
            setattr(self, f"ch{x}", ch)
        self.trigger = Trigger(self, "A")
        self.measurement = Measurement(self, 4)

    @property
    def sample_rate(self):
        """(float): sample rate in hertz
            The oscilloscope always acquires at this maximum rate and the displayed
            waveform is decimated according to the time per division setting.
        """
        return float(self._visa.query("HORIZONTAL:SAMPLERATE?"))

    @property
    def record_length(self):
        """value (int): 1e3, 1e4, 1e5, 1e6, 5e6, 1e7 samples"""
        return int(self._visa.query("HORIZONTAL:RECORDLENGTH?"))

    @record_length.setter
    def record_length(self, value):
        self._visa.write(f"HORIZONTAL:RECORDLENGTH {int(value)}")

    @property
    def horizontal_delay_mode(self):
        """value (str): delay_time or pretrigger_percent
            delay_time: center waveform horizontal_position seconds after trigger occurs
            pretrigger_percent: horizontal_position percent of record length
        """
        mode = self._visa.query("HORIZONTAL:DELAY:MODE?")
        return "delay_time" if mode in ["ON", "1"] else "pretrigger_percent"

    @horizontal_delay_mode.setter
    @validate
    def horizontal_delay_mode(self, value):
        mode = "ON" if value == "delay_time" else "OFF"
        self._visa.write(f"HORIZONTAL:DELAY:MODE {mode}")

    @property
    def horizontal_position(self):
        """value (float): position of waveform on display based on horizontal_delay_mode
            delay_time: center waveform horizontal_position seconds after trigger occurs
            pretrigger_percent: horizontal_position percent of record length
        """
        delay_time = float(self._visa.query("HORIZONTAL:DELAY:TIME?"))
        pretrigger_percent = float(self._visa.query("HORIZONTAL:POSITION?"))
        if self.horizontal_delay_mode in ["0", "OFF"]:
            horizontal_position = pretrigger_percent
        else:
            horizontal_position = delay_time
        return horizontal_position

    @horizontal_position.setter
    @validate
    def horizontal_position(self, value):
        if self.horizontal_delay_mode in ["0", "OFF"]:
            self._visa.write(f"HORIZONTAL:POSITION {value}")
        else:
            self._visa.write(f"HORIZONTAL:DELAY:TIME {value}")

    @property
    def acquisition_mode(self):
        """value (str): SAMPLE, PEAKDETECT, HIRES, AVERAGE, ENVELOPE
            AVERAGE - specify num_averages
            ENVELOPE - specify num_envelopes
        """
        return self._visa.query("ACQUIRE:MODE?")

    @acquisition_mode.setter
    @validate
    def acquisition_mode(self, value):
        self._visa.write(f"ACQUIRE:MODE {value}")

    def _get_wfmpre(self, inout=""):
        return super()._get_wfmpre("OUT")


class Channel(ChannelBase, kind="CH"):
    """Oscilloscope channel

    Attributes:
        owner (Instrument)
        CHx (int): channel number
    """

    def __init__(self, owner, CHx):
        super().__init__(owner, CHx)
        self._probe = Probe(self)

    @property
    def probe(self):
        """(Probe): probe"""
        return self._probe

    @property
    def label(self):
        """label (str): channel label"""
        return self._visa.query(f"CH{self._ch}:LABEL?").strip('"')

    @label.setter
    @validate
    def label(self, value):
        self._visa.write(f"CH{self._ch}:LABEL '{value}'")

    @property
    def coupling(self):
        """value (str): channel coupling AC or DC"""
        return self._visa.query(f"CH{self._ch}:COUPLING?")

    @coupling.setter
    @validate
    def coupling(self, value):
        self._visa.write(f"CH{self._ch}:COUPLING {value}")

    @property
    def termination(self):
        """value (float): channel termination in ohms, (1e6, 50, 75) Ω"""
        return float(self._visa.query(f"CH{self._ch}:TER?"))

    @termination.setter
    @validate
    def termination(self, value):
        self._visa.write(f"CH{self._ch}:TER {float(value)}")

    @property
    def bandwidth(self):
        """value (float or str): float hertz or FULL"""
        bandwidth = self._visa.query(f"CH{self._ch}:BANDWIDTH?")
        try:
            bandwidth = float(bandwidth)
        except ValueError:
            pass
        return bandwidth

    @bandwidth.setter
    @validate
    def bandwidth(self, value):
        try:
            bandwidth = float(value)
        except ValueError:
            bandwidth = value
        self._visa.write(f"CH{self._ch}:BANDWIDTH {bandwidth}")

    @property
    def trigger_level(self):
        """value (float or str): A-trigger threshold level, {voltage, ECL, TTL}
            float: level in volts
            str: ECL (-1.3 V) or TTL (1.4 V)
        """
        level = self._visa.query(f"TRIGGER:A:LEVEL:CH{self._ch}?")
        try:
            level = float(level)
        except ValueError:
            pass
        return level

    @trigger_level.setter
    @validate
    def trigger_level(self, value):
        self._visa.write(f"TRIGGER:A:LEVEL:CH{self._ch} {value}")

    @property
    def logic_input(self):
        """value (str): input condition for logic triggering, {HIGH, LOW, X}"""
        q_str = f"TRIGGER:A:LOGIC:INPUT:CH{self._ch}?"
        return self._visa.query(q_str)

    @logic_input.setter
    @validate
    def logic_input(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT:CH{self._ch} {value}"
        self._visa.write(cmd_str)

    @property
    def logic_threshold(self):
        """value (float or str): A-trigger logic threshold, {voltage, ECL, TTL}"""
        q_str = f"TRIGGER:A:LOGIC:THRESHOLD:CH{self._ch}?"
        threshold = self._visa.query(q_str)
        try:
            threshold = float(threshold)
        except ValueError:
            pass
        return threshold

    @logic_threshold.setter
    @validate
    def logic_threshold(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:THRESHOLD:CH{self._ch} {value}"
        self._visa.write(cmd_str)


class MeasurementSlot:
    """Automated measurement system of the oscilloscope

    Attributes:
        slot (int): measurement slot; 1 to 4
        resource (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, slot, resource):
        self._slot = slot
        self._source = "CH1"
        self._type = "amplitude"
        self._visa = resource
        self._visa.write(f"MEASUREMENT:MEAS{self._slot}:SOURCE1 {self._source}")
        self._visa.write(f"MEASUREMENT:MEAS{self._slot}:TYPE {self._type}")
        self._state = "OFF"
        self._visa.write(f"MEASUREMENT:MEAS{self._slot}:STATE {self._state}")

    @property
    def source(self):
        """value (str): source channel; CHx, x in 1 to 4"""
        self._source = self._visa.query(f"MEASUREMENT:MEAS{self._slot}:SOURCE?")
        return self._source

    @source.setter
    @validate
    def source(self, value):
        self._visa.write(f"MEASUREMENT:MEAS{self._slot}:SOURCE {value}")
        self._source = self._visa.query(f"MEASUREMENT:MEAS{self._slot}:SOURCE?")

    @property
    def measurement_type(self):
        """value (str): measurement type; amplitude, frequency, etc."""
        self._type = self._visa.query(f"MEASUREMENT:MEAS{self._slot}:TYPE?")
        return self._type

    @measurement_type.setter
    @validate
    def measurement_type(self, value):
        self._visa.write(f"MEASUREMENT:MEAS{self._slot}:TYPE {value}")
        self._type = self._visa.query(f"MEASUREMENT:MEAS{self._slot}:TYPE?")

    @property
    def state(self):
        """value (str): ON or OFF"""
        self._state = self._visa.query(f"MEASUREMENT:MEAS{self._slot}:STATE?")
        return self._state

    @state.setter
    @validate
    def state(self, value):
        self._visa.write(f"MEASUREMENT:MEAS{self._slot}:STATE {value}")
        self._state = self._visa.query(f"MEASUREMENT:MEAS{self._slot}:STATE?")

    @property
    def value(self):
        """(float): measurement value"""
        return float(self._visa.query(f"MEASUREMENT:MEAS{self._slot}:VALUE?"))

    @property
    def min(self):
        """(float): minimum measurement value"""
        return float(self._visa.query(f"MEASUREMENT:MEAS{self._slot}:MINIMUM?"))

    @property
    def max(self):
        """(float): maximum measurement value"""
        return float(self._visa.query(f"MEASUREMENT:MEAS{self._slot}:MAXIMUM?"))

    @property
    def avg(self):
        """(float): average measurement value"""
        return float(self._visa.query(f"MEASUREMENT:MEAS{self._slot}:MEAN?"))

    @property
    def stddev(self):
        """(float): standard deviation of measurement"""
        return float(self._visa.query(f"MEASUREMENT:MEAS{self._slot}:STDDEV?"))

    @property
    def unit(self):
        """(str): unit of measurement"""
        return self._visa.query(f"MEASUREMENT:MEAS{self._slot}:UNITS?")

    def __repr__(self):
        return str(self.value)


class Probe(ProbeBase, kind="Probe"):
    """Oscilloscope probe

    Attributes:
        owner (Instrument): Instrument instance
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    @property
    def model(self):
        """(str): probe model"""
        return self._visa.query(f"CH{self._ch}:PROBE:MODEL?").strip('"')

    @property
    def gain(self):
        """value (float): probe gain factor (output/input)"""
        return float(self._visa.query(f"CH{self._ch}:PROBE:GAIN?"))

    @gain.setter
    def gain(self, value):
        self._visa.write(f"CH{self._ch}:PROBE:GAIN {value}")

    @property
    def impedance(self):
        """value (float): probe impedance in ohms"""
        return float(self._visa.query(f"CH{self._ch}:PROBE:RESISTANCE?"))

    def __repr__(self):
        owner = self._owner.__repr__().strip("<>")
        return f"<{owner} {self._kind} {self.model}>"



"""Versioning"""
import argparse

__version__ = "1.0.16"


def main():
    """Set the version attribute from the cmd line parameters"""
    parser = argparse.ArgumentParser(description="Set __version__")
    parser.add_argument("major", type=int)
    parser.add_argument("minor", type=int)
    parser.add_argument("micro", type=int)
    args = parser.parse_args()
    version_str = f"{args.major}.{args.minor}.{args.micro}"
    with open(__file__) as f:
        lines = f.readlines()
    i = lines.index('__version__ = "0.0.0"\n')
    new_version = lines[i].replace("0.0.0", version_str)
    print(new_version)
    lines[i] = new_version
    with open(__file__, "wt") as f:
        f.writelines(lines)


if __name__ == "__main__":
    main()



"""Measurement subsystem"""
from unyt import unyt_quantity
from tekinstr.common import validate
from tekinstr.instrument import InstrumentSubsystem


class Measurement(InstrumentSubsystem, kind="Measurement"):
    """Automated measurement system of the oscilloscope

    Attributes:
        owner (Instrument)
        n_slots (int): numbers of measurement slots
    """

    def __init__(self, owner, n_slots):
        super().__init__(owner)
        self._n_slots = n_slots
        self._has_stats = self._has_stats_hw()
        self._slots = [MeasurementSlot(self, i) for i in range(1, n_slots + 1)]
        self._initialize()

    def _has_stats_hw(self):
        original_deser = int(self._visa.query("DESE?"))
        self._visa.write("DESE 16")  # check only for EXE bit
        self._visa.write("*CLS")
        self._visa.write("MEASUREMENT:MEAS1:MEAN?")
        self._visa.query("*ESR?")
        msgs = self._visa.query("EVENT?").split("\n")
        event_code = int(msgs[-1])
        self._visa.write(f"DESE {original_deser}")
        return event_code == 0

    def _initialize(self):
        for i in range(1, self._n_slots + 1):
            q_str = f"MEASUREMENT:MEAS{i}:"
            state = self._visa.query(q_str + "STATE?")
            if state in ["1", "ON"]:
                source = self._visa.query(q_str + "SOURCE?").lower()
                measurement_type = self._visa.query(q_str + "TYPE?").lower()
                measurement = self._slots[i - 1]
                setattr(self, source + "_" + measurement_type, measurement)

    def add(self, source, measurement_type):
        """Add a measurement

        Args:
            source (str): source channel; CHx
            measurement_type (str): type of measurement; amplitude, frequency, etc.
        """
        for slot in self._slots:
            if slot.state in ["0", "OFF"]:
                measurement = slot
                break
        else:
            raise ValueError("No measurement slots available. Remove a measurement")
        measurement.source = source
        measurement.measurement_type = measurement_type
        measurement.state = "ON"
        setattr(self, source + "_" + measurement_type, measurement)

    def remove(self, name):
        """Remove a measurement

        Args:
            name (str): measurement name
        """
        measurement = getattr(self, name, None)
        if measurement is None:
            raise ValueError(f"Unknown measurement {name}")
        measurement.state = "OFF"
        delattr(self, name)


class MeasurementSlot(InstrumentSubsystem, kind="MeasurementSlot"):
    """A measurement slot

    Attributes:
        owner (Measurement)
        slot (int): measurement slot
    """

    def __init__(self, owner, slot):
        super().__init__(owner)
        self._slot = slot
        self._has_stats = owner._has_stats

    @property
    def source(self):
        """value (str): source channel; CHx"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:SOURCE?"
        return self._visa.query(q_str)

    @source.setter
    @validate
    def source(self, value):
        cmd_str = f"MEASUREMENT:MEAS{self._slot}:SOURCE {value}"
        self._visa.write(cmd_str)

    @property
    def measurement_type(self):
        """value (str): measurement type; amplitude, frequency, etc."""
        q_str = f"MEASUREMENT:MEAS{self._slot}:TYPE?"
        return self._visa.query(q_str)

    @measurement_type.setter
    @validate
    def measurement_type(self, value):
        cmd_str = f"MEASUREMENT:MEAS{self._slot}:TYPE {value}"
        self._visa.write(cmd_str)

    @property
    def state(self):
        """value (str): {ON, OFF}"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:STATE?"
        return self._visa.query(q_str)

    @state.setter
    @validate
    def state(self, value):
        cmd_str = f"MEASUREMENT:MEAS{self._slot}:STATE {value}"
        self._visa.write(cmd_str)

    @property
    def value(self):
        """(float): measurement value"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:VALUE?"
        return float(self._visa.query(q_str))

    @property
    def min(self):
        """(float): minimum measurement value"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:MINIMUM?"
        return float(self._visa.query(q_str))

    @property
    def max(self):
        """(float): maximum measurement value"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:MAXIMUM?"
        return float(self._visa.query(q_str))

    @property
    def avg(self):
        """(float): average measurement value"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:MEAN?"
        return float(self._visa.query(q_str))

    @property
    def stddev(self):
        """(float): standard deviation of measurement"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:STDDEV?"
        return float(self._visa.query(q_str))

    @property
    def unit(self):
        """(str): unit of measurement"""
        q_str = f"MEASUREMENT:MEAS{self._slot}:UNITS?"
        return self._visa.query(q_str).strip('"')

    def __repr__(self):
        return str(unyt_quantity(self.value, self.unit))

    def __dir__(self):
        attrs = super().__dir__()
        for attr in ["source", "measurement_type", "state", "unit"]:
            attrs.remove(attr)
        if not self._has_stats:
            for attr in ["min", "max", "avg", "stddev"]:
                attrs.remove(attr)
        return attrs



"""Digital voltmeter (DVM) definition"""
from tekinstr.dvm import DVMBase


class DVM(DVMBase, kind="DVM"):
    """Digital voltmeter instrument

    Attributes:
        instr (Model)
        owner (Instrument)
        visa (pyvisa.resources.Resource): pyvisa resource
    """



"""LabVIEW's waveform data type in Python"""
from numbers import Number
import numpy as np
from unyt import matplotlib_support, unyt_array

matplotlib_support()
matplotlib_support.label_style = "/"


class WaveformDT(np.lib.mixins.NDArrayOperatorsMixin):
    """Python implementation of LabVIEW's waveform data type

    LabVIEW's waveform data type has three required attributes: t0, dt, and Y.
    Additional attributes can be set and are included in the returned WaveformDT.
    WaveformDT has the function to_xy() that will generate the x-axis array from the
    t0, dt and number of samples in the Y array.

    Attributes:
        Y (array-like): data
        dt (float): wf_increment
        t0 (float or datetime): wf_start_time

    Example:

        >>> data = WaveformDT([1,2,3], 1, 0)
        >>> x, y = data.to_xy()
        >>> x
        array([0., 1., 2.])
        >>> y
        array([1, 2, 3])

    WaveformDT supports a variety of workflows with and without units where unit
    support is provided by the Unyt library.

        >>> data.xunit = "s"
        >>> data.yunit = "V"
        >>> x, y = data.to_xy()
        >>> x
        unyt_array([0., 1., 2.], 's')
        >>> y
        unyt_array([0, 1, 2], 'V')

    WaveformDT supports Matplotlib and its labeled data interface:

        >>> import matplotlib.pyplot as plt
        >>> waveform = WaveformDT([1,2,3], 1, 0)
        >>> plt.plot('x', 'y', 'r-', data=waveform)
        [<matplotlib.lines.Line2D object ... >]
        >>> plt.show()

    Note:
        The x-axis array will be relative time by default. For absolute time, set the
        relative parameter to False.
    """

    def __init__(self, Y, dt, t0):
        self.Y = np.asanyarray(Y)
        self.dt = dt
        self.t0 = t0
        self._yunit = None
        self._xunit = None

    def __array__(self):
        return self.Y

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        args = []
        for input_ in inputs:
            if isinstance(input_, WaveformDT):
                args.append(input_.Y)
            else:
                args.append(input_)
        output = kwargs.pop("out", (None,))[0]
        if output is not None:
            if isinstance(output, WaveformDT):
                kwargs["out"] = output.Y
            else:
                kwargs["out"] = output
        arr_result = self.Y.__array_ufunc__(ufunc, method, *args, **kwargs)
        if arr_result is NotImplemented:
            return NotImplemented
        if method == "at":
            return None
        result = WaveformDT(arr_result, self.dt, self.t0) if output is None else output
        return result

    def set_attributes(self, **kwargs):
        """Set waveform attributes"""
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def size(self):
        """Return the number of elements in Y"""
        return self.Y.size

    @property
    def yunit(self):
        """units (str, Unit): Y-axis unit"""
        unit = self.Y.units if isinstance(self.Y, unyt_array) else self._yunit
        self._yunit = unit
        return unit

    @yunit.setter
    def yunit(self, units):
        try:
            self.Y.convert_to_units(units)
        except AttributeError:
            pass
        self._yunit = units

    @property
    def xunit(self):
        """units (str, Unit): X-axis unit, based on dt"""
        unit = self.dt.units if isinstance(self.dt, unyt_array) else self._xunit
        self._xunit = unit
        return unit

    @xunit.setter
    def xunit(self, units):
        try:
            self.dt.convert_to_units(units)
        except AttributeError:
            pass
        self._xunit = units

    @property
    def min(self):
        """(float): minimum value"""
        return np.min(self.Y)

    @property
    def max(self):
        """(float): maximum value"""
        return np.max(self.Y)

    @property
    def std(self):
        """(float): standard deviation"""
        return np.std(self.Y)

    def to_xy(self, relative=True, inunits=True):
        """Generate the (x, y) tuple

        Parameters:
            relative (bool): y is relative time if True, absolute if False
            inunits (bool): convert arrays to unyt_array if yunit and xunit are defined

        Returns:
            tuple: x, y arrays
        """
        y = self.Y
        y = y.flatten()
        dt = self.dt
        t0 = self.t0
        t0_offset = getattr(self, "wf_start_offset", 0.0)
        samples = y.size
        if relative:
            t0 = t0 if isinstance(t0, Number) else 0.0
            t0 = t0 + t0_offset
            x = np.linspace(t0, t0 + samples * dt, samples, False)
        else:
            t0 = np.datetime64(t0.astimezone().replace(tzinfo=None))
            t0_array = np.asarray([t0] * samples)
            dt = np.timedelta64(np.uint32(dt * 1e9), "ns")
            dt_array = np.asarray(
                [np.timedelta64(0, "ns")] + [dt] * (samples - 1)
            ).cumsum()
            x = t0_array + dt_array
        if inunits:
            if not isinstance(x, unyt_array) and self._xunit is not None:
                x = unyt_array(x, self._xunit)
            if not isinstance(y, unyt_array) and self._yunit is not None:
                y = unyt_array(y, self._yunit)
        else:
            x = x.v if isinstance(x, unyt_array) else x
            y = y.v if isinstance(y, unyt_array) else y
        return (x, y)

    def head(self, n=5):
        """Return first n samples of the waveform

        Args:
            n (int): number of samples to return

        Returns:
            WaveformDT: first n samples
        """
        return self[:n]

    def tail(self, n=5):
        """Return the last n samples of the waveform

        Args:
            n (int): number of samples to return

        Returns:
            WaveformDT: last n samples
        """
        start_offset = self.t0 if isinstance(self.t0, Number) else 0.0
        start_offset += getattr(self, "wf_start_offset", 0.0)
        start_offset += (self.size - n) * self.dt
        wf = self[-n:]
        setattr(wf, "wf_start_offset", start_offset)
        return wf

    def __dir__(self):
        inst_attr = list(filter(lambda k: not k.startswith("_"), self.__dict__.keys()))
        cls_attr = list(filter(lambda k: not k.startswith("_"), dir(self.__class__)))
        return inst_attr + cls_attr

    def __getitem__(self, key):
        if key in ["y", "Y"]:
            return self.Y
        if key in ["x", "X"]:
            return self.to_xy()[0]
        return self.__class__(self.Y[key], self.dt, self.t0)

    def __repr__(self):
        t0 = self.t0 if isinstance(self.t0, Number) else 0.0
        t0 += getattr(self, "wf_start_offset", 0.0)
        dt = self.dt
        rows = []
        if self.size < 50:
            for i, sample in enumerate(self.Y):
                rows.append(f"{t0 + i*dt:11.4e}\t{sample:11.4e}")
        else:
            for i, sample in enumerate(self[:5].Y):
                rows.append(f"{t0 + i*dt:11.4e}\t{sample:11.4e}")
            rows.append(" ...")
            t0 = t0 + (self.size - 5) * dt
            for i, sample in enumerate(self[-5:].Y):
                rows.append(f"{t0 + i*dt:11.4e}\t{sample:11.4e}")
        rows.append(f"Length: {self.size}")
        rows.append(f"t0: {self.t0}")
        rows.append(f"dt: {self.dt:11.4e}")
        return "\n".join(rows)



"""MDO3000 Series Oscilloscope"""
import re
from tekinstr.model import Model
from tekinstr.mdo3000.oscilloscope import Oscilloscope
from tekinstr.mdo3000.dvm import DVM
from tekinstr.mdo3000.spectrum_analyzer import SpectrumAnalyzer
from tekinstr.common import validate


class MDO3000(Model):
    """MDO3000 Series Oscilloscope class

    Attributes:
        resource (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, visa):
        super().__init__(visa)
        features = self._get_configuration()
        n_channels = features["ANALOG:NUMCHANNELS"]
        self.oscilloscope = Oscilloscope(self, n_channels)
        if features["DVM"]:
            self.dvm = DVM(self)
        if features["RF:NUMCHANNELS"] > 0:
            self.spectrum_analyzer = SpectrumAnalyzer(self)
        self._display = self._get_select()[0]

    @property
    def features(self):
        """(dict): model features"""
        return self._get_configuration()

    def _get_configuration(self):
        """Get the instrument configuration"""
        bool_features = [
            "ADVMATH",
            "AFG",
            "APPLICATIONS:POWER",
            "ARB",
            "AUXIN",
            "DVM",
        ]
        int_features = [
            "ANALOG:NUMCHANNELS",
            "DIGITAL:NUMCHANNELS",
            "NUMMEAS",
            "RF:NUMCHANNELS",
        ]
        configuration = {}
        for feature in bool_features:
            value = bool(int(self._visa.query(f"CONFIGURATION:{feature}?")))
            configuration[feature] = value
        for feature in int_features:
            value = int(self._visa.query(f"CONFIGURATION:{feature}?"))
            configuration[feature] = value
        return configuration

    def _get_select(self):
        """Get the current display selection"""
        original_header_state = self._visa.query("HEADER?")
        self._visa.write("HEADER ON")
        select_str = self._visa.query("SELECT?").replace(":SELECT:", "")
        raw_select = dict([kv.split(" ", maxsplit=1) for kv in select_str.split(";")])
        displayed_channels = set()
        control_channel = raw_select.pop("CONTROL", None)
        for k, v in raw_select.items():
            if v in ["1", "ON"]:
                k = "RF" if k.startswith("RF") else k
                displayed_channels.add(k)
        self._visa.write(f"HEADER {original_header_state}")
        return (displayed_channels, control_channel)

    @property
    def display(self):
        """channel(s) (str, list): display the given channel(s);
            'CHx', 'CHx:y', ['CHx', 'CHy'], or 'RF'
            The oscilloscope and spectrum analyzer are mutually exclusive systems and
            cannot be displayed simultaneously.
        """
        self._display = self._get_select()[0]
        return self._display

    @display.setter
    @validate
    def display(self, channels):
        chs = []
        if isinstance(channels, str) and channels.startswith("CH"):
            match = re.match("^CH(?P<first>[1-4])(?P<last>:[1-4])?$", channels)
            if match is None:
                raise ValueError(f"{channels} not a valid specification")
            first = int(match.group("first"))
            last = match.group("last")
            if last is None:
                last = first
            else:
                last = int(last[-1])
            for ch in range(first, last + 1):
                chs.append(f"CH{ch}")
            channels = chs
        elif isinstance(channels, str) and channels.startswith("RF"):
            channels = ["RF_NORMAL"]
        channels = set(channels)
        displayed_channels = self.display
        for channel in displayed_channels - channels:
            self._visa.write(f"SELECT:{channel} OFF")
        for channel in channels - displayed_channels:
            self._visa.write(f"SELECT:{channel} ON")
        self._display = self._get_select()[0]



"""Base Digital Voltmeter (DVM) definition"""
from tekinstr.common import validate
from tekinstr.instrument import Instrument


class DVMBase(Instrument, kind="DVMBase"):
    """Digital voltmeter instrument

    Attributes:
        instr (Model)
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    @property
    def mode(self):
        """value (str): ACRMS, ACDCRMS, DC, Frequency or OFF"""
        return self._visa.query("DVM:MODE?")

    @mode.setter
    @validate
    def mode(self, value):
        self._visa.write(f"DVM:MODE {value}")

    @property
    def source(self):
        """value (str): CHx"""
        return self._visa.query("DVM:SOURCE?")

    @source.setter
    @validate
    def source(self, value):
        self._visa.write(f"DVM:SOURCE {value}")

    @property
    def value(self):
        """(float): measurement value"""
        return float(self._visa.query("DVM:MEASUREMENT:VALUE?"))

    @property
    def min(self):
        """(float): minimum measurement value"""
        return float(self._visa.query("DVM:MEASUREMENT:HISTORY:MINIMUM?"))

    @property
    def max(self):
        """(float): maximum measurement value"""
        return float(self._visa.query("DVM:MEASUREMENT:HISTORY:MAXIMUM?"))

    @property
    def avg(self):
        """(float): average measurement value"""
        return float(self._visa.query("DVM:MEASUREMENT:HISTORY:AVERAGE?"))

    @property
    def frequency(self):
        """(float): frequency"""
        return float(self._visa.query("DVM:MEASUREMENT:FREQUENCY?"))



"""Tekinstr - the Pythonic way of communicating with Tektronix oscilloscopes"""
import pyvisa
from tekinstr.common import _get_idn
from tekinstr.mdo3000.mdo3000 import MDO3000
from tekinstr.tds3000.tds3000 import TDS3000
from tekinstr.mso4000b.mso4000b import MSO4000B
from tekinstr.version import __version__

MODEL_CLASS = {
    "MDO3012": MDO3000,
    "MDO3014": MDO3000,
    "MDO3022": MDO3000,
    "MDO3024": MDO3000,
    "MDO3032": MDO3000,
    "MDO3034": MDO3000,
    "MDO3052": MDO3000,
    "MDO3054": MDO3000,
    "MDO3102": MDO3000,
    "MDO3104": MDO3000,
    "TDS3064B": TDS3000,
    "MSO4104B": MSO4000B,
    "MSO4102B": MSO4000B,
    "MSO4014B": MSO4000B,
    "MSO4034B": MSO4000B,
    "MSO4054B": MSO4000B,
    "DPO4104B-L": MSO4000B,
    "MDO4104-6": MSO4000B,
}


class CommChannel:
    """Connect to a Tektronix oscilloscope using VISA

    Attributes:
        address (str): instrument's TCPIP address or host name

    Returns:
        (CommChannel or Model subclass)
    """

    def __init__(self, address):
        self._address = address
        self._rm = pyvisa.ResourceManager()
        self._visa = self._rm.open_resource(f"TCPIP::{address}")
        self._visa.read_termination = "\n"

    def __enter__(self):
        # self._visa.lock_excl()
        self._visa.write("*CLS")
        idn = _get_idn(self._visa)
        if idn.manufacturer != "TEKTRONIX":
            raise ValueError(f"Device at {self._address} is not a Tektronix model")
        try:
            return MODEL_CLASS[idn.model](self._visa)
        except KeyError:
            raise NotImplementedError(f"{idn.model} not currently supported") from None

    def __exit__(self, exc_type, exc_value, exc_tb):
        # self._visa.unlock()
        self._visa.close()
        self._rm.close()

    def get_instrument(self):
        """Return the instrument object"""
        return self.__enter__()

    def close(self):
        """Close the CommChannel"""
        self.__exit__(None, None, None)


__all__ = ["CommChannel"]



"""Common definitions"""
from collections import namedtuple
import functools

IDN = namedtuple("IDN", ["manufacturer", "model", "serial_number", "firmware_version"])


def _get_idn(visa):
    idn = visa.query("*IDN?").strip(":\n").split(",")
    manufacturer = idn[0]
    model = idn[1].replace(" ", "")
    serial_number = idn[2]
    versions = dict([field.split(":") for field in idn[3].strip().split(" ")])
    firmware_version = versions["FV"]
    return IDN(manufacturer, model, serial_number, firmware_version)


class CommandError(Exception):
    """Raised when CME bit of SESR is set"""


def validate(func):
    """Read the Command Error bit (CME) of the Standard Event Status Register (SESR)"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # pylint: disable=protected-access
        cme = 32
        self = args[0]
        self._visa.write("*CLS")
        original_deser = int(self._visa.query("DESE?"))
        self._visa.write(f"DESE {cme}")
        result = func(*args, **kwargs)
        if bool(int(self._visa.query("*ESR?")) & cme):
            msg = self._visa.query("EVMSG?").strip().split(",")[1].strip('"')
            self._visa.write(f"DESE {original_deser}")
            raise CommandError(msg)
        self._visa.write(f"DESE {original_deser}")
        return result

    return wrapper


class TekBase:
    """tekinstr base class

    Attributes:
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, visa):
        self._visa = visa

    def __setattr__(self, name, value):
        if hasattr(self, name) and isinstance(getattr(self, name), TekBase):
            raise AttributeError(f"can't set '{name}'")
        if hasattr(self.__class__, name):
            prop = getattr(self.__class__, name)
            try:
                prop.fset(self, value)
            except TypeError:
                raise AttributeError(f"can't set '{name}'") from None
        else:
            self.__dict__[name] = value

    def __repr__(self):
        raise NotImplementedError

    def __dir__(self):
        inst_attr = list(filter(lambda k: not k.startswith("_"), self.__dict__.keys()))
        cls_attr = list(filter(lambda k: not k.startswith("_"), dir(self.__class__)))
        return inst_attr + cls_attr



"""TDS3000 Series Oscilloscope"""
import re
from tekinstr.model import Model
from tekinstr.tds3000.oscilloscope import Oscilloscope


class TDS3000(Model):
    """TDS3000 Series Oscilloscope class

    Attributes:
        resource (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, visa):
        super().__init__(visa)
        self._visa.timeout = 4000
        pattern = "^TDS30(?P<full_bw>[1-6])(?P<n_channels>[24])[BC]*$"
        match = re.match(pattern, self.model)
        n_channels = int(match.group("n_channels"))
        self._full_bw = f"{match.group('full_bw')}00 MHz"
        self.oscilloscope = Oscilloscope(self, n_channels)

    @property
    def full_bandwidth(self):
        """(str): analog bandwidth when bandwidth is set to FULL"""
        return self._full_bw

    @property
    def battery_soc(self):
        """(float): percent battery state of charge"""
        soc = 100 * int(self._visa.query("POWER:BATTERY:GASGAUGE?")) / 15
        return soc



"""Trigger"""
from tekinstr.instrument import InstrumentSubsystem
from tekinstr.trigger import TriggerBase, LogicInput
from tekinstr.common import validate


class Trigger(TriggerBase, kind="Trigger"):
    """Trigger class

    Attributes:
        owner (Instrument)
        visa (pyvisa.resources.Resource): pyvisa resource
        designation (str): A or B
    """

    def __init__(self, owner, designation):
        super().__init__(owner, designation)
        self.edge = EdgeTrigger(self)
        if designation == "A":
            self.logic = LogicTrigger(self)
            self.pulse = PulseTrigger(self)

    def __dir__(self):
        attrs = super().__dir__()
        kind = self.kind
        if kind == "EDGE":
            attrs.remove("logic")
            attrs.remove("pulse")
        elif kind == "LOGIC":
            attrs.remove("edge")
            attrs.remove("pulse")
        else:
            attrs.remove("edge")
            attrs.remove("logic")
        return attrs


class EdgeTrigger(InstrumentSubsystem, kind="Edge"):
    """Edge trigger class

    Attributes:
        owner (Instrument)
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._designation = owner._designation

    def __repr__(self):
        instr_model = self._instr.model
        instr = self._owner._owner._kind
        subsys = self._owner._kind
        return f"<{instr_model} {instr} {self._designation}-{subsys}: {self._kind}>"

    @property
    def coupling(self):
        """value (str): {AC, DC, HFRej, LFRej, NOISErej}"""
        return self._visa.query(f"TRIGGER:{self._designation}:EDGE:COUPLING?")

    @coupling.setter
    @validate
    def coupling(self, value):
        self._visa.write(f"TRIGGER:{self._designation}:EDGE:COUPLING {value}")

    @property
    def slope(self):
        """value (str): trigger on slope RISE or FALL"""
        return self._visa.query(f"TRIGGER:{self._designation}:EDGE:SLOPE?")

    @slope.setter
    @validate
    def slope(self, value):
        self._visa.write(f"TRIGGER:{self._designation}:EDGE:SLOPE {value}")

    @property
    def source(self):
        """value (str): trigger source, CHx, EXT or LINE"""
        return self._visa.query(f"TRIGGER:{self._designation}:EDGE:SOURCE?")

    @source.setter
    @validate
    def source(self, value):
        self._visa.write(f"TRIGGER:{self._designation}:EDGE:SOURCE {value}")


class LogicTrigger(InstrumentSubsystem, kind="Logic"):
    """Logic trigger class

    Attributes:
        owner (Instrument)
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._designation = "A"
        self._input_1 = LogicInput(self, 1)
        self._input_2 = LogicInput(self, 2)

    def __repr__(self):
        instr_model = self._instr.model
        instr = self._owner._owner._kind
        subsys = self._owner._kind
        return f"<{instr_model} {instr} {self._designation}-{subsys}: {self._kind}>"

    @property
    def logic_class(self):
        """value (str): trigger class, {PATTERN, STATE}"""
        return self._visa.query("TRIGGER:A:LOGIC:CLASS?")

    @logic_class.setter
    @validate
    def logic_class(self, value):
        self._visa.write(f"TRIGGER:A:LOGIC:CLASS {value}")

    @property
    def input_1(self):
        """(LogicInput): logic trigger input 1 object"""
        return self._input_1

    @property
    def input_2(self):
        """(LogicInput): logic trigger input 2 object"""
        return self._input_2

    @property
    def deltatime(self):
        """value (float): pattern trigger delta time condition"""
        q_str = f"TRIGGER:A:LOGIC:PATTERN:DELTATIME?"
        return float(self._visa.query(q_str))

    @deltatime.setter
    @validate
    def deltatime(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:PATTERN:DELTATIME {value}"
        self._visa.write(cmd_str)

    @property
    def function(self):
        """value (str): logic function for pattern trigger, {AND, NAND, NOR, OR}"""
        q_str = f"TRIGGER:A:LOGIC:PATTERN:FUNCTION?"
        return self._visa.query(q_str)

    @function.setter
    @validate
    def function(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:PATTERN:FUNCTION {value}"
        self._visa.write(cmd_str)

    @property
    def when_pattern(self):
        """value (str): trigger condition
            PATTERN: TRUE, FALSE, LESSTHAN, MORETHAN, EQUAL, or NOTEQUAL
        """
        q_str = f"TRIGGER:A:LOGIC:PATTERN:WHEN?"
        return self._visa.query(q_str)

    @when_pattern.setter
    @validate
    def when_pattern(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:PATTERN:WHEN {value}"
        self._visa.write(cmd_str)

    @property
    def when_state(self):
        """value (str): trigger condition
            STATE: TRUE or FALSE
        """
        q_str = f"TRIGGER:A:LOGIC:STATE:WHEN?"
        return self._visa.query(q_str)

    @when_state.setter
    @validate
    def when_state(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:STATE:WHEN {value}"
        self._visa.write(cmd_str)

    def __dir__(self):
        attrs = super().__dir__()
        if self.logic_class == "STATE":
            for attr in ["deltatime", "function", "when_pattern"]:
                attrs.remove(attr)
        else:
            attrs.remove("when_state")
        return attrs


class PulseTrigger(InstrumentSubsystem, kind="Pulse"):
    """Pulse trigger class

    Attributes:
        owner (Instrument)
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._designation = "A"

    def __repr__(self):
        instr_model = self._instr.model
        instr = self._owner._owner._kind
        subsys = self._owner._kind
        return f"<{instr_model} {instr} {self._designation}-{subsys}: {self._kind}>"

    @property
    def pulse_class(self):
        """value (str): pulse trigger class {WIDTH, RUNT, SLEWRATE}"""
        q_str = f"TRIGGER:A:PULSE:CLASS?"
        return self._visa.query(q_str)

    @pulse_class.setter
    @validate
    def pulse_class(self, value):
        cmd_str = f"TRIGGER:A:PULSE:CLASS {value}"
        self._visa.write(cmd_str)

    @property
    def runt_polarity(self):
        """value (str): {EITHER, POSITIVE, NEGATIVE}"""
        q_str = f"TRIGGER:A:PULSE:RUNT:POLARITY?"
        return self._visa.query(q_str)

    @runt_polarity.setter
    @validate
    def runt_polarity(self, value):
        cmd_str = f"TRIGGER:A:PULSE:RUNT:POLARITY {value}"
        self._visa.write(cmd_str)

    @property
    def runt_threshold(self):
        """value (tuple or str): {(high, low), TTL, ECL}"""
        q_str = f"TRIGGER:A:PULSE:RUNT:THRESHOLD?"
        thresholds = self._visa.query(q_str).split(";")
        return tuple(float(t) for t in thresholds)

    @runt_threshold.setter
    @validate
    def runt_threshold(self, value):
        cmd_str = "TRIGGER:A:PULSE:RUNT:THRESHOLD:"
        if isinstance(value, tuple):
            high, low = value
            self._visa.write(cmd_str + f"HIGH {high}")
            self._visa.write(cmd_str + f"LOW {low}")
        else:
            self._visa.write(cmd_str + f"BOTH {value}")

    @property
    def runt_when(self):
        """value (str): {OCCURS, LESSTHAN, MOETHAN, EQUAL, NOTEQUAL}"""
        q_str = "TRIGGER:A:PULSE:RUNT:WHEN?"
        return self._visa.query(q_str)

    @runt_when.setter
    @validate
    def runt_when(self, value):
        cmd_str = f"TRIGGER:A:PULSE:RUNT:WHEN {value}"
        self._visa.write(cmd_str)

    @property
    def runt_width(self):
        """value (float): runt pulse width in seconds"""
        q_str = "TRIGGER:A:PULSE:RUNT:WIDTH?"
        return float(self._visa.query(q_str))

    @runt_width.setter
    @validate
    def runt_width(self, value):
        cmd_str = f"TRIGGER:A:PULSE:RUNT:WIDTH {value}"
        self._visa.write(cmd_str)

    @property
    def slewrate_deltatime(self):
        """value (float): slew rate trigger delta time condition"""
        q_str = f"TRIGGER:A:PULSE:SLEWRATE:DELTATIME?"
        return float(self._visa.query(q_str))

    @slewrate_deltatime.setter
    @validate
    def slewrate_deltatime(self, value):
        cmd_str = f"TRIGGER:A:PULSE:SLEWRATE:DELTATIME {value}"
        self._visa.write(cmd_str)

    @property
    def slewrate_polarity(self):
        """value (str): {EITHER, POSITIVE, NEGATIVE}"""
        q_str = f"TRIGGER:A:PULSE:SLEWRATE:POLARITY?"
        return self._visa.query(q_str)

    @slewrate_polarity.setter
    @validate
    def slewrate_polarity(self, value):
        cmd_str = f"TRIGGER:A:PULSE:SLEWRATE:POLARITY {value}"
        self._visa.write(cmd_str)

    @property
    def slewrate(self):
        """value (float): slew rate value in volts per second"""
        q_str = "TRIGGER:A:PULSE:SLEWRATE:SLEWRATE?"
        return float(self._visa.query(q_str))

    @slewrate.setter
    def slewrate(self, value):
        cmd_str = f"TRIGGER:A:PULSE:SLEWRATE:SLEWRATE {value}"
        self._visa.write(cmd_str)

    @property
    def slewrate_threshold(self):
        """value (tuple or str): {(high, low), TTL, ECL}"""
        q_str = f"TRIGGER:A:PULSE:SLEWRATE:THRESHOLD?"
        thresholds = self._visa.query(q_str).split(";")
        return tuple(float(t) for t in thresholds)

    @slewrate_threshold.setter
    @validate
    def slewrate_threshold(self, value):
        cmd_str = "TRIGGER:A:PULSE:SLEWRATE:THRESHOLD:"
        if isinstance(value, tuple):
            high, low = value
            self._visa.write(cmd_str + f"HIGH {high}")
            self._visa.write(cmd_str + f"LOW {low}")
        else:
            self._visa.write(cmd_str + f"BOTH {value}")

    @property
    def slewrate_when(self):
        """value (str): {OCCURS, LESSTHAN, MOETHAN, EQUAL, NOTEQUAL}"""
        q_str = "TRIGGER:A:PULSE:SLEWRATE:WHEN?"
        return self._visa.query(q_str)

    @slewrate_when.setter
    @validate
    def slewrate_when(self, value):
        cmd_str = f"TRIGGER:A:PULSE:SLEWRATE:WHEN {value}"
        self._visa.write(cmd_str)

    @property
    def width_polarity(self):
        """value (str): {POSITIVE, NEGATIVE}"""
        q_str = f"TRIGGER:A:PULSE:WIDTH:POLARITY?"
        return self._visa.query(q_str)

    @width_polarity.setter
    @validate
    def width_polarity(self, value):
        cmd_str = f"TRIGGER:A:PULSE:WIDTH:POLARITY {value}"
        self._visa.write(cmd_str)

    @property
    def width_when(self):
        """value (str): {LESSTHAN, MOETHAN, EQUAL, NOTEQUAL}"""
        q_str = "TRIGGER:A:PULSE:WIDTH:WHEN?"
        return self._visa.query(q_str)

    @width_when.setter
    @validate
    def width_when(self, value):
        cmd_str = f"TRIGGER:A:PULSE:WIDTH:WHEN {value}"
        self._visa.write(cmd_str)

    @property
    def pulse_width(self):
        """value (float): pulse width trigger time period in seconds"""
        q_str = "TRIGGER:A:PULSE:WIDTH:WIDTH?"
        return float(self._visa.query(q_str))

    @pulse_width.setter
    @validate
    def pulse_width(self, value):
        cmd_str = f"TRIGGER:A:PULSE:WIDTH:WIDTH {value}"
        self._visa.write(cmd_str)

    def __dir__(self):
        attrs = super().__dir__()
        width_attrs = ["width_polarity", "width_when", "pulse_width"]
        runt_attrs = ["runt_polarity", "runt_threshold", "runt_when", "runt_width"]
        slewrate_attrs = [
            "slewrate_deltatime",
            "slewrate_polarity",
            "slewrate",
            "slewrate_threshold",
            "slewrate_when",
        ]
        pulse_class = self.pulse_class
        if pulse_class == "WIDTH":
            for attr in runt_attrs + slewrate_attrs:
                attrs.remove(attr)
        elif pulse_class == "RUNT":
            for attr in width_attrs + slewrate_attrs:
                attrs.remove(attr)
        else:
            for attr in width_attrs + runt_attrs:
                attrs.remove(attr)
        return attrs
